//////////////////////////////////////
//------- DANGER DEBUG ONLY! -------//
//////////////////////////////////////

// Please check this area everytime you upload! 
bool DEBUG         = false;  // This must be true for any debug commands to work
bool debugTest     = false;  // Extra failsafe for the start up test ramp   
const int debugPwm = 700;   // Steady-state speed of the motor for debug testing. 

// To see all commands run under debug mode
// use Ctrl+F and search DEBUG ONLY.

//////////////////////////////////////
//------------Limits----------------//
//////////////////////////////////////

bool motorVoltageLimiting = true;  // False defaults to Current Limiting 

float startLimit        = 42;

float currentLimit1     = 29;
float currentLimit2     = 32;
float currentLimit3     = 35;
float voltageBypassLimit       = 37;

float currentLimit1     = 29;
float currentLimit2     = 32;
float currentLimit3     = 35;
float voltageBypassLimit       = 37;

//////////////////////////////////////
//----------------------------------//
//////////////////////////////////////

/////////////////////////////////////////////////////

/*       SD Setup variables and Functions          */

/////////////////////////////////////////////////////

#include "FS.h"
#include "SD.h"
#include "SPI.h"

unsigned long dataMillis = 0;
const int   dataInterval = 100;
unsigned long raceTime   = 0;
String dataString; 

  // SD Card command functions
  
void appendTitle(fs::FS &fs, const char * path, const char * message)
{
  File file = fs.open(path, FILE_APPEND);
  file.print(message);
  file.close();
}

void appendData(fs::FS &fs, const char * path, String  message)
{
  File file = fs.open(path, FILE_APPEND);
  file.print(message);
  file.close();
}

/////////////////////////////////////////////////////

/*              Motor Drive Variables              */

/////////////////////////////////////////////////////


#define PWMa 27
#define TH1 35
#define TH2 32
#define IN1 25
#define IN2 26

int throttle1 = 0;
int throttle2 = 0;
int input1   = 0;
int input2   = 0;

const int freq        = 15000;
const int resolution  = 10;
const int pwmChannela = 0;
int       pwm         = 0;
int       warp        = 0;
int       carMode     = 1;
int       driveMode   = 1;

bool startButtonTimer = true;
const int waitKeyInt  = 4000; 
unsigned long waitKey = 0;

bool firstTouch       = true;
bool rampInc          = false;
bool TFlag            = true;
int  TPos             = 0;  

float Kp              = 0.35;
int totalError        = 0;
int previousError     = 0;
float u;

/////////////////////////////////////////////////////

/*           Timer Setup variables               */

/////////////////////////////////////////////////////

unsigned long lowPriorityMillis = 0;
const int lowPriorityInterval   = 500;          

unsigned long medPriorityMillis = 0;
const int medPriorityInterval   = 250;

unsigned long hiPriorityMillis  = 0;
const int hiPriorityInterval    = 125;

/////////////////////////////////////////////////////

/*              ADS1115 Sensor Setup               */

/////////////////////////////////////////////////////

#include <Wire.h>
#include <Adafruit_ADS1X15.h>
Adafruit_ADS1115 ads;  /* Use this for the 16-bit version */
//const float gain = 0.0001875 // for Gain = 2/3x
const float gain = 0.000125; // for Gain = 1x
const float mapb = 4.907;

const int offsetSmoothing  = 60;
const int currentSmoothing = 3;
int       zeroingValue     = 0;

float batteryVoltage = 0;
float motorVoltage = 0;
float Current      = 0;
float convert      = 0;

float   ampHours = 0;
long    pasttime = 0;

  // The ADC input range (or gain) can be changed via the following
  // functions, but be careful never to exceed VDD +0.3V max, or to
  // exceed the upper and lower limits if you adjust the input range!
  // Setting these values incorrectly may destroy your ADC!
  //                                                                ADS1015  ADS1115
  //                                                                -------  -------
  // ads.setGain(GAIN_TWOTHIRDS);  // 2/3x gain +/- 6.144V  1 bit = 3mV      0.1875mV (default)
  // ads.setGain(GAIN_ONE);        // 1x gain   +/- 4.096V  1 bit = 2mV      0.125mV
  //ads.setGain(GAIN_TWO);         // 2x gain   +/- 2.048V  1 bit = 1mV      0.0625mV
  // ads.setGain(GAIN_FOUR);       // 4x gain   +/- 1.024V  1 bit = 0.5mV    0.03125mV
  // ads.setGain(GAIN_EIGHT);      // 8x gain   +/- 0.512V  1 bit = 0.25mV   0.015625mV
  // ads.setGain(GAIN_SIXTEEN);    // 16x gain  +/- 0.256V  1 bit = 0.125mV  0.0078125mV

  // In theory with a gain set to one, the ads1115 would see 0.5 amp as a 101.89bit increment.
  // therefore we can set the mapb (milli amps per bit value) as 4.907 (500ma/101.89bit)

/////////////////////////////////////////////////////

/*           Defining RPM Interrupt                */

/////////////////////////////////////////////////////

int RPM = 0;
const uint8_t rpmPin = 33;
volatile unsigned long motorPoll = 0;
unsigned long lastMotorSpeedPollTime = 0;

void IRAM_ATTR isr() {
  motorPoll++;
}


/////////////////////////////////////////////////////

/*           Defining Telemetry                    */

/////////////////////////////////////////////////////

byte DataToSend = 0;

/////////////////////////////////////////////////////


/////////////////////////////////////////////////////


/*                      SETUP                      */


/////////////////////////////////////////////////////

/////////////////////////////////////////////////////

void setup()
{  
  Serial.begin(115200);

  // Setup Driver Inputs
  pinMode(TH1, INPUT);
  pinMode(TH2, INPUT);
  pinMode(IN1, INPUT);
  pinMode(IN2, INPUT);

  
  //pinMode(input3, INPUT);
  //pinMode(input4, INPUT);

  // RPM Interrupt
  attachInterrupt(rpmPin, isr, RISING);


  // Setup SD card 
  if(!SD.begin(5)){
    Serial.println("Card Mount Failed");
  }
  appendTitle(SD, "/HadesData.txt", "TIME, BV, MV, C, RPM, PWM , AH, MODE, TPos");


  // Start ADS1115 ADC module
  if (!ads.begin())
  {
    Serial.println("Failed to initialize ADS.");
    while (1);
  } 
  ads.setGain(GAIN_ONE);

  // ADC Zeroing
  int adcAverage = 0;
  for (int i = 0; i<offsetSmoothing; i++)
  { 
    adcAverage = adcAverage + ads.readADC_SingleEnded(2);
  }
  
  zeroingValue = adcAverage/offsetSmoothing;

  // initiate soft start checking
  waitKey = millis();  


  ledcSetup(pwmChannela, freq, resolution);
  ledcAttachPin(PWMa, pwmChannela);
//////////////////////////////////////
//------- DANGER DEBUG ONLY! -------//
//////////////////////////////////////

  if(DEBUG == true){
  /* Attach the LED PWM Channel to the GPIO Pin */
    delay(2000);
    
    ledcWrite(pwmChannela, 0);
        for(int i = 0; i<debugPwm; i++){
        delay(30);
        ledcWrite(pwmChannela, i);
      }
  }

//////////////////////////////////////
//----------------------------------//
//////////////////////////////////////

}

/////////////////////////////////////////////////////

/////////////////////////////////////////////////////


/*                  Main Loop                      */


/////////////////////////////////////////////////////

/////////////////////////////////////////////////////


void loop(){

  // Get Drive Inputs
  throttle1 = digitalRead(TH1);
  throttle2 = digitalRead(TH2);
  input1 = digitalRead(IN1);
  input2 = digitalRead(IN2);
  getSensorData();
  timer();

  //Serial.println("--------------------------");
  //Serial.println(throttle1);
  //Serial.println(throttle2);
  //Serial.println(input1);
  //Serial.println(input2);


  // Check if Start procedure is active
  if((millis() > (waitKey + waitKeyInt)) && Current <= 2 && motorVoltage <= 3.5){
    driveMode = 0;
  }
  else if(motorVoltage >= 21.5){
    driveMode = carMode;
  }

  // Mode 4 = override all other modes and start functions ("BYPASS")
  if(carMode == 4){
    driveMode = carMode;
  }

  // Main Control Loop 
  if(throttle1 == LOW || throttle2 == LOW)
  {
    TPos = 1;
    startButtonTimer = true;
    switch(driveMode){
    
      case 0: 
        startMode();
        break;

      case 1:
        if(voltageLimiting == true){
          pwm += PIDu(voltageLimit1);        
        }
        else{
          pwm += PIDu(currentLimit1);        
        }
        

        break;

      case 2:
        pwm += PIDu(currentLimit2);
        break;

      case 3:
        pwm += PIDu(currentLimit3);
        break;
        
      case 4:
        pwm += PIDu(bypassLimit);
        break;
    }
  }
  if(throttle1 == HIGH || throttle2 == HIGH)
  {
   Serial.println("-------------");
   Serial.println("OFF Loop");
    if (TFlag == true){
      TPos = 1;
    }
    else if(TFlag == false){
      TPos = 0;
    }

    pwm = 0;
    firstTouch = true;
    
    if(startButtonTimer == true){
      startButtonTimer = false;
      waitKey = millis();    
    }   
  }
  if(pwm>=1023){
    pwm = 1023;
  }
  else if(pwm < 0){
    pwm = 0;
  }
  
  // DEBUG ONLY
  if(DEBUG == true){
    pwm = debugPwm;
  }
  Serial.println(pwm);
  ledcWrite(pwmChannela, pwm);
}

/////////////////////////////////////////////////////

/*              Drive Control Functions            */

/////////////////////////////////////////////////////

void startMode(){ 
  Serial.println("-------------");
  Serial.println("Start Loop");
  if(firstTouch == true){
    firstTouch = false; 
    pwm = 350;
  }
  
  if(rampInc == true && Current <= startLimit){
    rampInc = false;
    pwm = pwm + 20;
  }
  
}

int PIDu(int limit){
  Serial.println("-------------");
  Serial.println("PIDu Loop");
  float error = limit - Current;
  float u1;
  u1 = Kp * error; 
  return u1;
}

/////////////////////////////////////////////////////

/*              Sensor Data Function               */

/////////////////////////////////////////////////////

//returns nothing, uses global variable ;(.
void getSensorData()
{

  int16_t adc0, adc1, adc2, adc3;
  float volts0, volts1, volts2;
  

  adc0 = ads.readADC_SingleEnded(0);
  adc1 = ads.readADC_SingleEnded(1);
  adc2 = ads.readADC_SingleEnded(2);

  adc2 -= zeroingValue;

  adc2 = constrain(adc2,0,30000);

  volts0 = adc0 * gain;
  volts1 = adc1 * gain;
  volts2 = adc2 * mapb;

  batteryVoltage  = (volts1 * 9.06) + 0.15;
  motorVoltage    = batteryVoltage - (volts0 * 9.06);
  Current         = volts2 / 1000;

  motorVoltage = constrain(motorVoltage,0,30);

  if(motorPoll >= 28){
    readMotorRPM();
  }
}

/////////////////////////////////////////////////////

/*              Amp Hours Function                 */

/////////////////////////////////////////////////////
void readAmpHours()
{ 
  long nowtime = millis();
  float difference = nowtime - pasttime;
  pasttime = nowtime;
  
  float Atime = (Current*1000) * difference;
  
  float Ahours = (Atime/1000)/3600;
  
  ampHours = ampHours + Ahours;
  //Serial.println(ampHours);
  convert = ampHours / 1000;
  DataToSend = convert/0.5;
  //DataToSend = 21.44;
}

/////////////////////////////////////////////////////

/*               Timed Tasks Function              */

/////////////////////////////////////////////////////
// Seperate function for timers
void timer()                                                                  // Seperate function for timers
{

  if (millis() >= medPriorityMillis + medPriorityInterval){
    medPriorityMillis += medPriorityInterval;
    rampInc = true;
    
  }
  
  if (millis() >= lowPriorityMillis + lowPriorityInterval){
    lowPriorityMillis += lowPriorityInterval;
    
    int tempMode = carMode;

    if((input1 == HIGH) && (input2 == HIGH)){
      carMode = 1;
    }
    else if((input1 == HIGH) && (input2 == LOW)){
      carMode = 2;
    }
    else if((input1 == LOW) && (input2 == HIGH)){
      carMode = 3;
    }
    else if((input1 == LOW) && (input2 == LOW)){
      carMode = 4;
    }

    Wire.beginTransmission(6); // transmit to device #4
    //Serial.println(DataToSend);
    Wire.write(DataToSend);    // sends one byte  
    Wire.endTransmission();    // stop transmitting    


    // DEBUG ONLY 
    if(DEBUG == true){

      Serial.println("-----------------------------------------------------------");

      Serial.print("BVolts: "); Serial.print(batteryVoltage); Serial.println("V");
      Serial.print("MVolts: "); Serial.print(motorVoltage); Serial.println("V");
      Serial.print("MVolts: "); Serial.print(Current); Serial.println("A");
      Serial.print("RPM: "); Serial.print(RPM);
      Serial.print("TH1: "); Serial.println(throttle1);
      Serial.print("TH2: "); Serial.println(throttle2);
      Serial.print("IN1: "); Serial.println(input1);
      Serial.print("IN2: "); Serial.println(input2);

    }


  }
  
  if (millis() >= hiPriorityMillis + hiPriorityInterval)  
  {
    hiPriorityMillis += hiPriorityInterval;
    saveData();
    TFlag = false; 
  }

}

/////////////////////////////////////////////////////

/*          Motor RPM Calculation Function         */

/////////////////////////////////////////////////////

void readMotorRPM()
{
 
  int tempMotorPoll = motorPoll;                                                    // Records the number of interrupts that have occured since the last function recall  
  motorPoll = 0;                                                                    // Reset previous value
  
  unsigned long tempLastMotorPollTime = lastMotorSpeedPollTime;                     // Use variables to remember the time when the last RPM was calculated.
  
  lastMotorSpeedPollTime = millis();                                                // Record the current time to be used in the this cycle and the next.
  float motorRevolutions = tempMotorPoll / 3;                                       // Now calculate the number of revolutions of the motor shaft
  
  motorRevolutions = motorRevolutions * 60.0;                                       // Calculate the revolutions in one minute
  float timeDifference = (lastMotorSpeedPollTime - tempLastMotorPollTime)/1000.0;   // Find the time difference in seconds
    
  float ShaftRPM = motorRevolutions / timeDifference;                               // Finally calculate RPM 
  RPM = ShaftRPM;
  Serial.println(RPM);
  
}

/////////////////////////////////////////////////////

/*          SD Card Save Function                  */

/////////////////////////////////////////////////////

void saveData()
{
  raceTime = millis();

  String (dataString) = "\n" + String(raceTime) + "," + String(batteryVoltage) + "," + String(motorVoltage) + "," + String(Current) + "," + String(RPM) + "," + String(pwm) + "," + String(convert) + "," + String(driveMode) + "," + String(TPos);
  appendData(SD, "/HadesData.txt", dataString);
  
}
