/////////////////////////////////////////////////////

/*           Discharge Characteristics             */

/////////////////////////////////////////////////////

// user variables

bool quickCycle               = false;
bool parallel                 = false;
const long quickCycleDuration = 0;              // delay in milliseconds

const float dischargeLimit = 11;                     // single battery votlage limit
const int   numOfBatteries = 2;                      // Number of 12V batteries in series
const int dischargeRate    = 230;                    // discharge rate pwm value 0 - 255
const int maxCurrent       = 15; 

int rampRate = 2;                                     // value between 1 and 5

// private variables
int count                 = 0;
int startRamp             = 1; 
int ramp                  = 10;
unsigned long  rampMillis = 0;
const int    rampInterval = 500;
float     calculatedLimit = 0;   

unsigned long quickCycleTimer = 0;        
              

/////////////////////////////////////////////////////

/*       SD Setup variables and Functions          */

/////////////////////////////////////////////////////

#include "FS.h"
#include "SD.h"
#include "SPI.h"

unsigned long dataMillis = 0;
const int   dataInterval = 1000;
unsigned long raceTime   = 0;

unsigned long serialMillis = 0;
const int    serialInterval = 2000;

String dataString; 

  // SD Card command functions
  
void appendTitle(fs::FS &fs, const char * path, const char * message)
{
  File file = fs.open(path, FILE_APPEND);
  file.print(message);
  file.close();
}

void appendData(fs::FS &fs, const char * path, String  message)
{
  File file = fs.open(path, FILE_APPEND);
  file.print(message);
  file.close();
}

/////////////////////////////////////////////////////

/*           PWM and Throttle Setup                */

/////////////////////////////////////////////////////

#define PWMa    27
const int freq         = 15000;
const int resolution   = 8;
const int pwmChannela  = 0;
int       pwm          = 0;



/////////////////////////////////////////////////////

/*              ADS1115 Sensor Setup               */

/////////////////////////////////////////////////////

#include <Wire.h>
#include <Adafruit_ADS1X15.h>
Adafruit_ADS1115 ads;  /* Use this for the 16-bit version */
//const float gain = 0.0001875 // for Gain = 2/3x
const float gain = 0.000125; // for Gain = 1x
const float mapb = 4.907;

const int offsetSmoothing  = 60;
const int currentSmoothing = 3;
int       zeroingValue     = 0;

float batteryVoltage = 0;
float motorVoltage = 0;
float Current      = 0;
float aH           = 0;

float   ampHours = 0;
long    pasttime = 0;

  // The ADC input range (or gain) can be changed via the following
  // functions, but be careful never to exceed VDD +0.3V max, or to
  // exceed the upper and lower limits if you adjust the input range!
  // Setting these values incorrectly may destroy your ADC!
  //                                                                ADS1015  ADS1115
  //                                                                -------  -------
  // ads.setGain(GAIN_TWOTHIRDS);  // 2/3x gain +/- 6.144V  1 bit = 3mV      0.1875mV (default)
  // ads.setGain(GAIN_ONE);        // 1x gain   +/- 4.096V  1 bit = 2mV      0.125mV
  //ads.setGain(GAIN_TWO);         // 2x gain   +/- 2.048V  1 bit = 1mV      0.0625mV
  // ads.setGain(GAIN_FOUR);       // 4x gain   +/- 1.024V  1 bit = 0.5mV    0.03125mV
  // ads.setGain(GAIN_EIGHT);      // 8x gain   +/- 0.512V  1 bit = 0.25mV   0.015625mV
  // ads.setGain(GAIN_SIXTEEN);    // 16x gain  +/- 0.256V  1 bit = 0.125mV  0.0078125mV

  // In theory with a gain set to one, the ads1115 would see 0.5 amp as a 101.89bit increment.
  // therefore we can set the mapb (milli amps per bit value) as 4.907 (500ma/101.89bit)   


/////////////////////////////////////////////////////

/////////////////////////////////////////////////////


/*                      SETUP                      */


/////////////////////////////////////////////////////

/////////////////////////////////////////////////////

void setup() 
{
  Serial.begin(115200);                                                         // Any baud rate lower than 115200 is inadequate 

  // Start ADS1115 ADC module
  if (!ads.begin())
  {
    Serial.println("Failed to initialize ADS.");
    while (1);
  } 
  ads.setGain(GAIN_ONE);

  // Setup battery characterstics
  calculatedLimit = dischargeLimit*numOfBatteries;
  
  // Set PWM Channel and frequency
  ledcSetup(pwmChannela, freq, resolution);
  ledcAttachPin(PWMa, pwmChannela);                                           // Attach the LED PWM Channel to the GPIO Pin 
  ledcWrite(pwmChannela, 0);                                                     // Set initial pwm value to 0 so no drive is active

  // ADC Zeroing
  int adcAverage = 0;
  for (int i = 0; i<offsetSmoothing; i++)
  { 
    adcAverage = adcAverage + ads.readADC_SingleEnded(2);
  } 
  zeroingValue = adcAverage/offsetSmoothing;

  // Setup SD card 
  if(!SD.begin(5)){
    Serial.println("Card Mount Failed");
    //while(1);
  }
  appendTitle(SD, "/HadesData.txt", "TIME, BV, MV, C, pwm, AH");

  delay(6000);

  if(startRamp == 1){
    Serial.println("---------------------------");
    Serial.println("SYSTEM: RAMP INITIATED     ");
    Serial.println("Discharge Limit: ");
    Serial.println(calculatedLimit);
    Serial.println("---------------------------");
  }

  quickCycleTimer = millis() + quickCycleDuration;
}

/////////////////////////////////////////////////////

/////////////////////////////////////////////////////


/*                  Main Loop                      */


/////////////////////////////////////////////////////

/////////////////////////////////////////////////////

void loop() 
{ 

  // Get Sensor Variables
  getSensorData();
  timer();

  
  

  if(startRamp == 1){
      if (millis() >= rampMillis + rampInterval){
        rampMillis += rampInterval;
        ramp += rampRate;
        pwm = ramp;
      }
  }
  else{
    pwm = dischargeRate;
  }
  
  if(startRamp == 1 && ramp >= dischargeRate){
    startRamp = 0;
    pwm = dischargeRate;
    Serial.println("SYSTEM: RAMP COMPLETE      ");
    Serial.print("SYSTEM: HOLDING AT -  ");
    Serial.println(dischargeRate);
    Serial.println("---------------------------");
  }

  if(Current >= maxCurrent){
    pwm = 0; 
    Serial.println("ERROR: CURRENT MAX EXCEEDED");
    Serial.println("---------------------------");
  }
  
  if(batteryVoltage < calculatedLimit){
    count++;
    if(count >= 10)
    {
      pwm = 0; 
      Serial.println("SYSTEM: DISCHARGE COMPLETE");
      Serial.println("---------------------------");
      saveData();
      updateSerial();
      ledcWrite(pwmChannela, pwm);
      while(1); 
    }
  }


// For quick cylce = true only
    if(parallel == true){
      if(dischargeRate >= 125){
        pwm = 0;
        Serial.println("SYSTEM: RATE TOO HIGH");
        Serial.println("---------------------------");
        saveData();
        updateSerial();
        ledcWrite(pwmChannela, pwm);        
        while(1);
      }
    }
    if(quickCycle == true){
      if(millis() >= quickCycleTimer)
      {
        pwm = 0; 
        Serial.println("SYSTEM: Timer COMPLETE");
        Serial.println("---------------------------");
        saveData();
        updateSerial();
        ledcWrite(pwmChannela, pwm);
        while(1); 
      }   
    }
    if(aH >= 30){
      if(millis() >= quickCycleTimer)
      {
        pwm = 0; 
        Serial.println("SYSTEM: Timer COMPLETE");
        Serial.println("---------------------------");
        saveData();
        updateSerial();
        ledcWrite(pwmChannela, pwm);
        while(1); 
      }   
    }


  if(pwm > dischargeRate){
    pwm = dischargeRate;
  }
  ledcWrite(pwmChannela, pwm);
  
}

/////////////////////////////////////////////////////

/////////////////////////////////////////////////////


/*                 Functions                       */             // These are the blocks of code that allow data to be gathered, read and written for all the boards functions
                                                                  

/////////////////////////////////////////////////////

/////////////////////////////////////////////////////

/////////////////////////////////////////////////////

/*              Sensor Data Function               */

/////////////////////////////////////////////////////

//returns nothing, uses global variable ;(.
void getSensorData()
{

  // 
  int16_t adc0, adc1, adc2, adc3;
  float volts0, volts1, volts2;
  
  // Get single shot sample from ADS sensor.
  adc0 = ads.readADC_SingleEnded(0);
  adc1 = ads.readADC_SingleEnded(1);
  adc2 = ads.readADC_SingleEnded(2);

  // Calibrate each reading.
  adc2 -= zeroingValue;

  // constrain output so there are no erroneous negative values
  adc2 = constrain(adc2,0,30000);

  // Multiply each reading with a calibrated fudge factor
  volts0 = adc0 * gain;
  volts1 = adc1 * gain;
  volts2 = adc2 * mapb; // mapb = milli amps per bit = 4.907

  // fudge the voltage readings and format to the correct scale. 
  batteryVoltage  = (volts1 * 9.06) + 0.15;
  motorVoltage    = batteryVoltage - (volts0 * 9.06);
  Current         = volts2 / 1000; //convert milli amps to amps

  motorVoltage = constrain(motorVoltage,0,30);



  readAmpHours();
}

/////////////////////////////////////////////////////

/*              Amp Hours Function                 */

/////////////////////////////////////////////////////
void readAmpHours()
{ 
  long nowtime = millis();
  float difference = nowtime - pasttime;
  pasttime = nowtime;
  
  float Atime = (Current*1000) * difference;
  
  float Ahours = (Atime/1000)/3600;
  
  ampHours = ampHours + Ahours;
  //Serial.println(ampHours);
  aH = ampHours / 1000;
  
  //DataToSend = 21.44;
}





void timer()                                                                  // Seperate function for timers
{
  if (millis() >= serialMillis + serialInterval)
  {
    serialMillis += serialInterval;
    updateSerial();
  }

  if (millis() >= dataMillis + dataInterval)  
  {
    dataMillis += dataInterval;
    getSensorData();
    saveData();
  }
}

void updateSerial()
{
  float watts = Current*motorVoltage;
  Serial.print("          OV: ");
  Serial.println(motorVoltage);
  Serial.print("          BV: ");
  Serial.println(batteryVoltage);
  Serial.print("           C: ");
  Serial.println(Current);
  Serial.print("          Ah: ");
  Serial.println(aH);
  Serial.print("       WATTS: ");
  Serial.println(watts);
  Serial.println("---------------------------");
}

void saveData()
{
  raceTime = millis();

  String (dataString) = "\n" + String(raceTime) + "," + String(batteryVoltage) + "," + String(motorVoltage) + "," + String(Current) + "," + String(pwm) + "," + String(aH);
  appendData(SD, "/HadesData.txt", dataString);

}
