#define throttle1 17
#define throttle2 18
uint8_t dispSetting = 0;
int state = 0;


/////////////////////////////////////////////////////

/*           Defining Button Interrupts            */

/////////////////////////////////////////////////////


struct Button {                                       //Struct Button is a function that creates a list of variables under a single name  
  const uint8_t pin;                                  //This makes it easier to setup multiple variable values on one command as shown below               
  bool pressed;
};

Button disp1 = {16,false};                           // Define disp1 onto GPIO 4   (Calling Button before a variable name now allows values to be assigned to pin and pressed(state))
Button disp2 = {4 ,false};                           // Define disp2 onto GPIO 0   
Button disp3 = {0 ,false};                           // Define disp3 onto GPIO 8

void IRAM_ATTR isr1()                               // Attach 3 seperate interrupts so that each button has a steady state bool variable that stops debouncing 
{
  disp1.pressed = true;                             // Have a bool toggle true/false (See dispChange() function) allows all switch bounces to be eliminated 
}                                                   // however, this relies upon the void loop() function running with 0 delays
void IRAM_ATTR isr2()
{
  disp2.pressed = true;
}
void IRAM_ATTR isr3()
{
  disp3.pressed = true;
}

void setup() 
{                                  
  Serial.begin(115200);

  pinMode(throttle1, INPUT);
  pinMode(disp1.pin, INPUT);
  pinMode(disp2.pin, INPUT);
  pinMode(disp3.pin, INPUT);

  attachInterrupt(disp1.pin, isr1, FALLING);
  attachInterrupt(disp2.pin, isr2, FALLING);
  attachInterrupt(disp3.pin, isr3, FALLING);

}

void loop() 
{
  state = digitalRead(throttle1);
  dispSetting = dispChange();
  Serial.println(dispSetting);

  switch (dispSetting)
  {
    case 1:
      Serial.println("display button 1 pressed");
      break;

    case 2:
      Serial.println("display button 2 pressed");
      break;

    case 3:
      Serial.println("display button 3 pressed");
      break;

    default:
      Serial.println("No Change");
      break; 
  }
  delay(1000);
}

uint8_t dispChange ()
{
  if(disp1.pressed == true)
  {
    disp1.pressed =false;
    return 1; 
  }
  else if(disp2.pressed == true)
  {
    disp2.pressed = false;
    return 2;
  }
  else if(disp3.pressed == true)
  {
    disp3.pressed = false;
    return 3;
  }
  else{
    return 0;
  }

}
